@@ -2,7 +2,7 @@ import vflow
 from vflow.convert import to_tuple, to_list
 
 
-class TestBasic():
+class TestBasic:
     def setup(self):
         self.pipeline = vflow.PCSPipeline()
         self.module_set = vflow.Vset(name='s', modules={})
@@ -14,8 +14,8 @@ class TestBasic():
         assert self.module is not None
 
     def test_iteration(self):
-        '''Tests that iterating over pipeline is same as iterating over its steps
-        '''
+        """Tests that iterating over pipeline is same as iterating over its steps
+        """
         self.pipeline.steps = [0, 1, 2]
         assert self.pipeline.steps[0] == 0
         assert self.pipeline[0] == 0, 'accessing pipeline steps'
@@ -24,8 +24,8 @@ class TestBasic():
         assert self.pipeline[1:] == [1, 2], 'slicing pipeline'
 
     def test_list_packing(self):
-        '''Test that packing / unpacking lists works appropriately
-        '''
+        """Test that packing / unpacking lists works appropriately
+        """
         start = [[0, 10], [1, 11], [2, 12]]
         X, y = to_tuple(start)
         packed = to_list((X, y))
